"""
This modul is there for the basic functions of all query's
"""
from abc import ABC, abstractmethod
from datetime import timedelta
from typing import Union, Tuple

from mariadb_sqlbuilder.helpful.validator import Validator
from mariadb_sqlbuilder.exepetions import BetweenValueIsBigger
from mariadb_sqlbuilder.helpful.arithmetic import Arithmetic


# get the name of a table column
def _get_tcn(tb, column: str) -> str:
    if tb.validator:
        tb.validator.check_column_exists(tb.table, column)
    return tb.table + "." + column


def _get_tcn_without_validator(table: str, column: str) -> str:
    return table + "." + column


def _get_tcn_validator(table: str, column: str, validator: Validator) -> str:
    if validator:
        validator.check_column_exists(table, column)
    return table + "." + column


def _transform_value_valid(value: Union[str, int, bool, ]) -> str:
    if value is None:
        return "NULL"
    elif isinstance(value, bool):
        return "TRUE" if value else "FALSE"
    elif isinstance(value, (int, Arithmetic)):
        return str(value)
    elif isinstance(value, timedelta):
        if "," in str(value):
            return f"'{str(value).split(', ')[1]}'"
        return f"'{value}'"
    return f"'{value}'"


class BaseBuilder(ABC):
    """
    TODO: add a description
    This is a dummy docstring.
    """

    def __init__(self, tb, **kwargs):
        self._tb = tb

    @abstractmethod
    def get_sql(self) -> str:
        """
        Get the sql script that was generated
        :return:
        """

    @property
    def tb(self):
        """
        returns the table builder
        :return:
        """
        return self._tb


class ConditionsBuilder(BaseBuilder):
    """
    TODO: add a description
    This is a dummy docstring.
    """

    def __init__(self, tb, **kwargs):
        super().__init__(tb)
        if "condition" in kwargs:
            self.__conditions = kwargs["condition"].get_conditions()
            self.__default_condition = kwargs["condition"].get_default_condition()
        else:
            self.__conditions = []
            self.__default_condition = "AND"

    def where(self, expression: Union[str, Arithmetic],
              value: Union[str, int, float], filter_operator: str = "="):
        """
        Adds a WHERE condition for an exact match of a column value.
        :param expression: a Column or an Arithmetic
        :param value:
        :param filter_operator:
        :return:
        """
        self.__check_if_or_and()
        if isinstance(expression, str):
            self.__conditions.append(f"{_get_tcn(self.tb, expression)} {filter_operator} "
                                          f"{_transform_value_valid(value)}")
        else:
            self.__conditions.append(f"{expression} {filter_operator} "
                                     f"{_transform_value_valid(value)}")
        return self

    def where_in(self, expression: Union[str, Arithmetic],
                 checked_list: Tuple[str, int, float]):
        """
        Adds a WHERE condition for a list of checked values in a column.
        :param expression: a Column or an Arithmetic
        :param checked_list:
        :return:
        """
        self.__check_if_or_and()
        if isinstance(expression, str):
            self.__conditions.append(
                f"{_get_tcn(self.tb, expression)} IN {str(checked_list)}"
            )
        else:
            self.__conditions.append(
                f"{expression} IN {str(checked_list)}"
            )
        return self

    def where_in_not(self, expression: Union[str, Arithmetic],
                     checked_list: Tuple[str, int, float]):
        """
        Adds a WHERE condition for a list of unchecked values in a column.
        :param expression: a Column or an Arithmetic
        :param checked_list:
        :return:
        """
        self.__check_if_or_and()
        if isinstance(expression, str):
            self.__conditions.append(
                f"{_get_tcn(self.tb, expression)} NOT IN {str(checked_list)}"
            )
        else:
            self.__conditions.append(
                f"{expression} NOT IN {str(checked_list)}"
            )
        return self

    def like(self, expression: Union[str, Arithmetic], value: Union[str, int, float]):
        """
        Adds a WHERE condition for a partial match of a column value.
        :param expression: a Column or an Arithmetic
        :param value:
        :return:
        """
        self.__check_if_or_and()
        if isinstance(expression, str):
            self.__conditions.append(
                f"{_get_tcn(self.tb, expression)} LIKE {_transform_value_valid(value)}"
            )
        else:
            self.__conditions.append(
                f"{expression} LIKE {_transform_value_valid(value)}"
            )
        return self

    def like_not(self, expression: Union[str, Arithmetic], value: Union[str, int, float]):
        """
        Adds a WHERE condition for a partial mismatch of a column value.
        :param expression: a Column or an Arithmetic
        :param value:
        :return:
        """
        self.__check_if_or_and()
        if isinstance(expression, str):
            self.__conditions.append(
                f"{_get_tcn(self.tb, expression)} NOT LIKE {_transform_value_valid(value)}"
            )
        else:
            self.__conditions.append(
                f"{expression} NOT LIKE {_transform_value_valid(value)}"
            )
        return self

    def between(self, expression: Union[str, Arithmetic],
                value1: Union[str, int], value2: Union[str, int, float]):
        """
        Adds a WHERE condition for a range of values in a column.
        :param expression: a Column or an Arithmetic
        :param value1:
        :param value2:
        :return:
        """
        self.__check_if_or_and()
        if isinstance(value1, (int, float)) and isinstance(value2, (int, float)):
            if value1 > value2:
                raise BetweenValueIsBigger("Value 1 is bigger then value 2")
        if isinstance(expression, str):
            self.__conditions.append(
                f"{_get_tcn(self.tb, expression)} "
                f"BETWEEN {_transform_value_valid(value1)} "
                f"AND {_transform_value_valid(value2)}"
            )
        else:
            self.__conditions.append(
                f"{expression} "
                f"BETWEEN {_transform_value_valid(value1)} "
                f"AND {_transform_value_valid(value2)}"
            )
        return self

    def between_not(self, expression: Union[str, Arithmetic],
                    value1: Union[str, int, float], value2: Union[str, int, float]):
        """
        Adds a WHERE condition for a range of values not in a column.
        :param expression: a Column or an Arithmetic
        :param value1:
        :param value2:
        :return:
        """
        self.__check_if_or_and()
        if isinstance(value1, (int, float)) and isinstance(value2, (int, float)):
            if value1 > value2:
                raise BetweenValueIsBigger("Value 1 is bigger then value 2")
        if isinstance(expression, str):
            self.__conditions.append(
                f"{_get_tcn(self.tb, expression)} "
                f"NOT BETWEEN {_transform_value_valid(value1)} "
                f"AND {_transform_value_valid(value2)}"
            )
        else:
            self.__conditions.append(
                f"{expression} "
                f"NOT BETWEEN {_transform_value_valid(value1)} "
                f"AND {_transform_value_valid(value2)}"
            )
        return self

    def is_not_null(self, column: str):
        """
        Adds a WHERE condition for a column that is not NULL.
        :param column:
        :return:
        """
        self.__check_if_or_and()
        self.__conditions.append(f"{_get_tcn(self.tb, column)} IS NOT NULL")
        return self

    def is_null(self, column: str):
        """
        Adds a WHERE condition for a column that is NULL.
        :param column:
        :return:
        """
        self.__check_if_or_and()
        self.__conditions.append(f"{_get_tcn(self.tb, column)} IS NULL")
        return self

    def is_true(self, column: str):
        """
        Adds a WHERE condition for a column that is True.
        :param column:
        :return:
        """
        self.__check_if_or_and()
        self.__conditions.append(f"{_get_tcn(self.tb, column)} IS TRUE")
        return self

    def is_not_true(self, column: str):
        """
        Adds a WHERE condition for a column that is not True.
        :param column:
        :return:
        """
        self.__check_if_or_and()
        self.__conditions.append(f"{_get_tcn(self.tb, column)} IS NOT TRUE")
        return self

    def is_false(self, column: str):
        """
        Adds a WHERE condition for a column that is False.
        :param column:
        :return:
        """
        self.__check_if_or_and()
        self.__conditions.append(f"{_get_tcn(self.tb, column)} IS FALSE")
        return self

    def is_not_false(self, column: str):
        """
        Adds a WHERE condition for a column that is not False.
        :param column:
        :return:
        """
        self.__check_if_or_and()
        self.__conditions.append(f"{_get_tcn(self.tb, column)} IS NOT FALSE")
        return self

    def __check_if_or_and(self):
        """
        Checks if the last added condition is either "AND" or "OR"
        and adds the default condition if not.
        :return:
        """
        if not self.__conditions:
            return
        if self.__conditions[-1] not in ["AND", "OR"]:
            self.__conditions.append(self.__default_condition)

    def OR(self):
        """
        Adds an OR condition to the WHERE conditions.
        :return:
        """
        if not self.__conditions or self.__conditions[-1] == "OR":
            return self
        elif self.__conditions[-1] == "AND":
            self.__conditions.pop(-1)
        self.__conditions.append("OR")
        return self

    def AND(self):
        """
        Adds an AND condition to the WHERE conditions.
        :return:
        """
        if not self.__conditions or self.__conditions[-1] == "AND":
            return self
        elif self.__conditions[-1] == "OR":
            self.__conditions.pop(-1)
        self.__conditions.append("AND")
        return self

    def default_and(self):
        """
        Sets the default condition to "AND".
        :return:
        """
        self.__default_condition = "AND"
        return self

    def default_or(self):
        """
        Sets the default condition to "OR".
        :return:
        """
        self.__default_condition = "OR"
        return self

    def get_default_condition(self):
        """
        Returns the default condition.
        :return:
        """
        return self.__default_condition

    def get_conditions(self):
        """
        Returns a copy of the current WHERE conditions list.
        :return:
        """
        return self.__conditions.copy()

    def _get_where_sql(self) -> str:
        """
        Returns the WHERE clause SQL statement.
        :return:
        """
        if not self.__conditions:
            return ""
        return "WHERE " + " ".join(self.__conditions)

    def is_conditions(self) -> bool:
        """
        Returns if conditions or not
        :return:
        """
        return bool(self.__conditions)

    @abstractmethod
    def get_sql(self) -> str:
        """
        Get the sql script that was generated
        :return:
        """



"""
Validator is a connector used class for
checking if a request possible, or it will throw errors

It checks

- if the table exists
- if the column exists
- if the value type the required type of the column

"""
import re
from datetime import datetime, timedelta
from typing import List, Tuple
from uuid import UUID

from _decimal import Decimal

from mariadb_sqlbuilder.exepetions import InvalidColumnType, ValidatorType,\
    ValidatorLength, ValidatorUnknown, ValidatorRange, ValidatorTableNotFound,\
    ValidatorColumnNotFound


class _Column:
    """
    This class contains all need data for checking if a value
    possible to add it in the database
    """

    def __init__(self, name: str, data_type: str, length: int, nullable: bool,
                 unsigned: bool, character_set_name: str):
        self.name = name
        self.nullable = nullable
        self.range: Tuple[int, int] = (0, 0)
        self.length = length
        self.unsigned = unsigned
        self.data_type_name = data_type
        self.data_type: any = None
        self.character_set_name = character_set_name
        if data_type in ["varchar", "text"]:
            self.data_type = str
        elif data_type in ["int", "bigint", "smallint", "mediumint", "tinyint", "year"]:
            self.data_type = int
            if data_type == "int":
                if unsigned:
                    if length < 10:
                        self.range = (0, int("9" * length))
                    else:
                        self.range = (0, 4294967295)
                else:
                    if length < 10:
                        self.range = (-int("9" * length), int("9" * length))
                    else:
                        self.range = (-2147483648, 2147483647)
            elif data_type == "bigint":
                if unsigned:
                    if length < 20:
                        self.range = (0, int("9" * length))
                    else:
                        self.range = (0, 18446744073709551615)
                else:
                    if length < 20:
                        self.range = (-int("9" * length), int("9" * length))
                    else:
                        self.range = (-9223372036854775808, 9223372036854775807)
            elif data_type == "smallint":
                if unsigned:
                    if length < 5:
                        self.range = (0, int("9" * length))
                    else:
                        self.range = (0, 65535)
                else:
                    if length < 5:
                        self.range = (-int("9" * length), int("9" * length))
                    else:
                        self.range = (-32768, 32767)
            elif data_type == "mediumint":
                if unsigned:
                    if length < 8:
                        self.range = (0, int("9" * length))
                    else:
                        self.range = (0, 16777215)
                else:
                    if length < 7:
                        self.range = (-int("9" * length), int("9" * length))
                    else:
                        self.range = (-8388608, 8388607)
            elif data_type == "tinyint":
                if unsigned:
                    if length < 3:
                        self.range = (0, int("9" * length))
                    else:
                        self.range = (0, 255)
                else:
                    if length < 3:
                        self.range = (-int("9" * length), int("9" * length))
                    else:
                        self.range = (-128, 127)
            elif data_type == "year":
                self.range = (4, 4)
        elif data_type in ["decimal"]:
            self.data_type = Decimal
        elif data_type in ["float", "double"]:
            self.data_type = float
            # I do not check it because they are to many
            # variables of the size of a float or double
            # https://mariadb.com/kb/en/float/
            # https://mariadb.com/kb/en/double/
        elif data_type in [
            "blob", "mediumblob", "longblob",
            "char", "inet4", "inet6",
            "mediumtext", "longtext", "text",
            "tinytext", "varchar", "tinyblob"
        ]:
            self.data_type = str
            if data_type == "text":
                if length is 0:
                    self.length = 65535
            elif data_type == "tinytext":
                if length is 0:
                    self.length = 255
            elif data_type == "mediumtext":
                if length is 0:
                    self.length = 16777215
            elif data_type == "longtext":
                if length is 0:
                    self.length = 4294967295
            elif data_type == "char":
                if length == 0:
                    self.length = 255
            elif data_type == "varchar":
                if length == 0:
                    if "utf8" in character_set_name:
                        self.length = 21844
                    else:
                        self.length = 65532
            elif data_type == "blob":
                if length == 0:
                    self.length = 65535
            elif data_type == "mediumblob":
                if length == 0:
                    self.length = 16777215
            elif data_type == "longblob":
                if length == 0:
                    self.length = 4294967295
            elif data_type == "tinyblob":
                if length == 0:
                    self.length = 255
        elif data_type in ["uuid"]:
            self.data_type = UUID
        elif data_type in ["date", "datetime", "timestamp"]:
            self.data_type = datetime
        elif data_type in ["time"]:
            self.data_type = timedelta
        elif data_type in ["enum"]:
            # Why i do not a checking
            # https://mariadb.com/kb/en/enum/
            self.data_type = list
        else:
            raise InvalidColumnType(
                f"The column '{name}' has an unsupported validator type -> {data_type}"
            )

    def check(self, value: any) -> bool:
        """
        Check if a value match the type requirements of the column
        :param value:
        :return:
        """
        if value is None and not self.nullable:
            raise ValidatorType("The given value is None, but the accepted"
                                "value can't be None")
        if value is None:
            return True
        if not isinstance(value, self.data_type):
            raise ValidatorType(
                f"The type ({type(value)}) of the given value do "
                f"not match the column type {self.data_type}."
            )
        if self.data_type == str:
            value: str
            if self.data_type_name in [
                "text", "tinytext", "mediumtext", "longtext", "varchar", "char"
            ]:
                length = len(value)
                if length > self.length:
                    raise ValidatorLength(
                        f"The length of the string ({length}) "
                        f"is bigger then the allowed length of {self.length}"
                    )
                return True
            elif self.data_type_name in [
                "blob", "tinyblob", "mediumblob", "longblob"
            ]:
                length = len(value.encode(self.character_set_name))
                if length > self.length:
                    raise ValidatorLength(
                        f"The bytes of the string ({length}) "
                        f"is bigger then the allowed bytes of {self.length}"
                    )
                return True
            elif self.data_type_name in ["inet4", "inet6"]:
                return True
            raise ValidatorUnknown("Oh... there is an internal problem with the type system")
        elif self.data_type == int:
            value: int
            if not self.range[0] <= value <= self.range[1]:
                raise ValidatorRange(
                    f"Your number ({str(value)}) is not in the range of the supported type {str(self.range)}."
                )
            return True
        elif self.data_type == float:
            return True
        elif self.data_type == UUID:
            return True
        elif self.data_type == datetime:
            return True
        elif self.data_type == timedelta:
            return True
        elif self.data_type == Decimal:
            return True
        elif self.data_type == list and self.name == "enum":
            if not all(isinstance(val, str) for val in value):
                raise ValidatorType(
                    "An enum only accept strings"
                )
            return True
        raise ValidatorUnknown("Internal error with the column types, 404 not found")


class Validator:
    """
    This class is for checking the existing of tables, columns and the
    right value types
    """

    def __init__(self, conn):
        self._conn = conn
        tables: List[List[str]] = conn.execute_fetchmany("SHOW tables;")
        tables: List[str] = [table[0] for table in tables]
        self.__structure = {}
        table: str
        table_join = "'" + "', '".join(tables) + "'"
        columns = conn.execute_fetchmany(
            "SELECT COLUMN_NAME, COLUMN_TYPE, IS_NULLABLE, CHARACTER_SET_NAME, TABLE_NAME " +
            "FROM information_schema.COLUMNS " +
            f"WHERE TABLE_NAME IN ({table_join}) and TABLE_SCHEMA = N'{self._conn.schema}';"
        )
        for column in columns:
            if not column[4] in self.__structure:
                self.__structure[column[4]] = {}
            data_type = column[1].split("(")[0] \
                if "(" in column[1] else column[1].split(" ")[0]
            match = re.search('\((\d+)\)', column[1])
            if match:
                length = int(match.group(1))
            else:
                length = 0
            unsigned = True if "unsigned" in column[1] else False
            self.__structure[column[4]][column[0]] = _Column(
                column[0], data_type, length, column[2], unsigned, column[3]
            )

    def check_table_exists(self, table: str):
        """
        Check if the table exists, if not it raise an error
        :param table:
        :return:
        """
        if table in self.__structure:
            return
        raise ValidatorTableNotFound(
            f"Table {table} not found. Check upper- and lowercase. If you change the database "
            "while the system is running, regenerate the validator"
        )

    def check_column_exists(self, table: str, column: str):
        """
        Check if the table and the column exists,
        if not it raise an error.
        :param table:
        :param column:
        :return:
        """
        self.check_table_exists(table)
        if column in self.__structure[table]:
            return
        raise ValidatorColumnNotFound(
            f"The column {column} in the table {table} not found. "
            f"Check upper- and lowercase. If you change the database "
            "while the system is running, regenerate the validator"
        )

    def check_value_type(self, table: str, colum: str, value: any):
        """
        Check if the given value the correct accepted type of the column,
        its also check if the table and column exists.
        If not it raise an error.
        :param table:
        :param colum:
        :param value:
        :return:
        """
        self.check_column_exists(table, colum)
        return self.__structure[table][colum].check(value)

    @property
    def structure(self):
        """
        Return the dictionary of the current structure.
        Helpful vor debugging
        :return:
        """
        return self.__structure.copy()



"""
This modul is there for build sql joins for your query

Maria DB SQL joins
https://mariadb.com/kb/en/joining-tables-with-join-clauses/

"""
from abc import ABC, abstractmethod

from .base_builder import _get_tcn_without_validator


class _JoinBuilder(ABC):
    """
    TODO: add a description
    This is a dummy docstring.
    """

    def __init__(self, table: str):
        self.from_table = ''
        self.table: str = table

    @abstractmethod
    def get_sql(self) -> str:
        """
        Abstract method that should be implemented by the subclasses.
        :return:
        """


class BaseJoinExtension:
    """
    TODO: add a description
    This is a dummy docstring.
    """

    def __init__(self, tb, **kwargs):
        self._tb = tb
        self._join_builders = []
        self._joins = []

    def join(self, join_builder: _JoinBuilder):
        """
        Adds a join statement to the query.
        :param join_builder:
        :return:
        """
        join_builder.from_table = self._tb.table
        self._tb.validator.check_table_exists(join_builder.table)
        if isinstance(join_builder, _ConditionsBuilder):
            join_builder: _ConditionsBuilder
            for condition in join_builder.conditions:
                self._tb.validator.check_column_exists(self._tb.table, condition[0])
                self._tb.validator.check_column_exists(join_builder.table, condition[1])
        self._join_builders.append(join_builder)
        self._joins.append(join_builder.get_sql())
        return self


class CrossJoinBuilder(_JoinBuilder):
    """
    TODO: add a description
    This is a dummy docstring.
    """

    def __init__(self, table: str, **kwargs):
        """
        Adds a join statement to the query.
        :param table:
        :param kwargs:
        """
        super().__init__(table)

    def get_sql(self) -> str:
        """
        Builds the SQL query for the CROSS JOIN statement.
        :return:
        """
        return f"CROSS JOIN {self.table} "


class _ConditionsBuilder(_JoinBuilder):
    """
    TODO: add a description
    This is a dummy docstring.
    """

    def __init__(self, table: str):
        super().__init__(table)
        self.conditions = []

    def condition(self, column_from_table: str, column_join_table: str):
        """
        Adds a join condition to the query.
        :param column_from_table:
        :param column_join_table:
        :return:
        """
        self.conditions.append([column_from_table, column_join_table])
        return self

    @abstractmethod
    def get_sql(self) -> str:
        """
        Abstract method that should be implemented by the subclasses.
        :return:
        """


class InnerJoinBuilder(_ConditionsBuilder):
    """
    TODO: add a description
    This is a dummy docstring.
    """

    def get_sql(self) -> str:
        """
        Builds the SQL query for the INNER JOIN statement.
        :return:
        """
        conditions = []
        for con in self.conditions:
            conditions.append(
                f"{_get_tcn_without_validator(self.from_table, con[0])} = " +
                f"{_get_tcn_without_validator(self.table, con[1])}"
            )
        return f"INNER JOIN {self.table} ON {' AND '.join(conditions)} "


class LeftJoinBuilder(_ConditionsBuilder):
    """
    TODO: add a description
    This is a dummy docstring.
    """

    def get_sql(self) -> str:
        """
        Returns a string representation of the left join operation between the two tables.
        :return:
        """
        conditions = []
        for con in self.conditions:
            conditions.append(
            f"{_get_tcn_without_validator(self.from_table, con[0])} = " +
            f"{_get_tcn_without_validator(self.table, con[1])}")
        return f"Left JOIN {self.table} ON {' AND '.join(conditions)} "


class RightJoinBuilder(_ConditionsBuilder):
    """
    TODO: add a description
    This is a dummy docstring.
    """

    def get_sql(self) -> str:
        """
        Returns a string representation of the right join operation between the two tables.
        :return:
        """
        conditions = []
        for con in self.conditions:
            conditions.append(
                f"{_get_tcn_without_validator(self.from_table, con[0])} = " +
                f"{_get_tcn_without_validator(self.table, con[1])}"
            )
        return f"RIGHT JOIN {self.table} ON {' AND '.join(conditions)} "



"""
This modul is there for build a sql update query
"""
from datetime import timedelta, datetime
from json import dumps
from typing import Union, Dict, List

from _decimal import Decimal

from mariadb_sqlbuilder.helpful.arithmetic import Arithmetic
from .base_builder import ConditionsBuilder, _transform_value_valid, \
    _get_tcn_without_validator
from .join_builder import BaseJoinExtension


class UpdateBuilder(ConditionsBuilder, BaseJoinExtension):
    """
    TODO: add a description
    This is a dummy docstring.
    """

    def __init__(self, tb, **kwargs):
        ConditionsBuilder.__init__(self, tb, **kwargs)
        BaseJoinExtension.__init__(self, tb, **kwargs)
        # check if variable already exists, else init it
        self.__toSet = {}
        self.sure_not_use_conditions = False
        self.__subSets = []
        self.__jsonBuildings = []

    def set(self, column, value: Union[str, int, float, bool,
        Arithmetic, timedelta, datetime, Decimal]):
        """
        Set a value for a column in the table to update.
        :param column:
        :param value:
        :return:
        """
        self.tb.validator.check_value_type(self.tb.table, column, value)
        self.__toSet[_get_tcn_without_validator(self.tb, column)] = _transform_value_valid(value)
        return self

    def join_set(self, join_table: str, join_column: str, value: Union[str, int, float, bool,
        Arithmetic, timedelta, datetime, Decimal]):
        """
        Set a value for a column in a join table.
        :param join_table:
        :param join_column:
        :param value:
        :return:
        """
        self.tb.validator.check_value_type(join_table, join_column, value)
        self.__toSet[
            _get_tcn_without_validator(join_table, join_column)
        ] = _transform_value_valid(value)
        return self

    def im_sure_im_not_use_conditions(self, im_sure: bool = True):
        """
        Set a flag to indicate that the where conditions are not needed.
        :param im_sure:
        :return:
        """
        self.sure_not_use_conditions = im_sure
        return self

    def execute(self):
        """
        Execute the update statement.
        :return:
        """
        if not self.is_conditions() and not self.sure_not_use_conditions:
            raise PermissionError('Update Builder: You are not sure enough not to use where')
        cursor = self.tb.connector.get_available_cursor()
        cursor.execute(
            self.get_sql()
        )
        if self.__subSets:
            for subset in self.__subSets:
                cursor.execute(subset.get_sql())
        cursor.connection.commit()
        self.tb.connector.release_cursor(cursor)

    def get_sql(self) -> str:
        """
        Get the SQL statement to execute.
        :return:
        """
        for element in self.__jsonBuildings:
            self.__set_json(element[0], element[1])
        sql = f"UPDATE {self.tb.table} " \
              f"{' '.join(self._joins) if self._joins else ''} " \
              f"SET " \
              f"{', '.join([f'{key} = {value}' for (key, value) in self.__toSet.items()])} " \
              f"{self._get_where_sql()};"
        return sql

    def __set_json(self, json: Dict[str, any], pop: List[str] = None):
        """
        Set values from a json object.
        :param json:
        :param pop:
        :return:
        """
        if pop is None:
            pop = []
        key: str
        value: any
        join_keys = [x.table for x in self._join_builders]
        for key, value in json.items():
            if isinstance(value, dict):
                if key in join_keys and not key in pop:
                    for sub_key, sub_value in value.items():
                        self.join_set(key, sub_key, sub_value)
                else:
                    self.set(key, dumps(value))
            else:
                self.set(key, value)
        return self

    def set_json(self, json: Dict[str, any], pop: List[str] = None):
        """
        Set values with a json object.
        :param json: dict with data example from select
        :param pop: pop keys from the json,
        if you have keys inside that are not a table but a dict/list
        :return:
        """
        self.__jsonBuildings.append([json, pop])
        return self



"""
Arithmetic Operators, math in sql

More information under
https://mariadb.com/kb/en/arithmetic-operators/
"""
from typing import Union


_invalid_arithmetic = ["+", "-", "*", "/", "%"]


class ArithmeticColumn:

    def __init__(self, table: str, column: str):
        self._column = table + "." + column

    def __str__(self):
        return self._column


def _check_type(variable: Union[ArithmeticColumn, 'Arithmetic', int, float]):
    if not isinstance(variable, (ArithmeticColumn, Arithmetic, int, float)):
        raise TypeError(
            "Variable is not a valid type, accept ArithmeticColumn, Arithmetic, int and float"
        )


class Arithmetic:

    def __init__(self, variable: Union[ArithmeticColumn, 'Arithmetic', int, float]):
        if not isinstance(variable, (ArithmeticColumn, int, float)):
            raise
        self.arithmetic_str = str(variable)

    def add(self, variable: Union[ArithmeticColumn, 'Arithmetic', int, float]):
        _check_type(variable)
        self.arithmetic_str += f" + {variable}"
        return self

    def sub(self, variable: Union[ArithmeticColumn, 'Arithmetic', int, float]):
        _check_type(variable)
        self.arithmetic_str += f" - {variable}"
        return self

    def mul(self, variable: Union[ArithmeticColumn, 'Arithmetic', int, float]):
        _check_type(variable)
        self.arithmetic_str += f" * {variable}"
        return self

    def div(self, variable: Union[ArithmeticColumn, 'Arithmetic', int, float]):
        _check_type(variable)
        self.arithmetic_str += f" / {variable}"
        return self

    def mod(self, variable: Union[ArithmeticColumn, 'Arithmetic', int, float]):
        _check_type(variable)
        self.arithmetic_str += f" % {variable}"
        return self

    def __str__(self):
        return f"({self.arithmetic_str})"



"""
This modul is there for fake a connection to a database,
if you only want to use the sql query builder
"""
from mariadb_sqlbuilder.helpful.validator_dummy import ValidatorDummy
from mariadb_sqlbuilder.builder import TableBuilder


class Connector:
    """
    TODO: add a description
    This is a dummy docstring.
    """

    def __init__(self, info_print: bool = True):
        if info_print:
            print("\n\033[36mMariaDB_SQLBuilder connect class is an dummy class.\n"
                  "\033[1;31mExecute and fetching function raise errors!\n"
                  "\033[93mMany functions of connect are not available.\n"
                  "get_sql() is still usable, read the wiki for more information." + "\033[0;0m\n")
        self.__validator = ValidatorDummy()

    def table(self, name: str) -> TableBuilder:
        """
        Get the functions for building the sql in the selected table
        :param name:
        :return:
        """
        return TableBuilder(self, name)

    def get_available_cursor(self):
        """
        This is a "dead end" for the execution functions. This will raise a NotImplementedError,
        which is helpful for a better assignment of the error.
        :return:
        """
        raise NotImplementedError("This function is only available with a active connection")

    @property
    def schema(self) -> str:
        """
        It's only a dummy function,
        its return's the schema name "MariaDBSQLBuilder"
        :return:
        """
        return "MariaDBSQLBuilder"

    @property
    def validator(self) -> ValidatorDummy:
        """
        Returns the current used validator
        :return:
        """
        return self.__validator



"""
This modul is there for save conditions for multiple uses
"""
from mariadb_sqlbuilder.builder.base_builder import ConditionsBuilder


class _DummyTB:
    """
    TODO: add a description
    This is a dummy docstring.
    """

    def __init__(self, table: str):
        self.__table = table

    @property
    def table(self):
        """
        Returns the dummy table name
        :return:
        """
        return self.__table


class ConditionsSaver(ConditionsBuilder):
    """
    TODO: add a description
    This is a dummy docstring.
    """

    def __init__(self, table: str, **kwargs):
        super().__init__(_DummyTB(table), **kwargs)

    def get_sql(self) -> str:
        """
        Get the sql script that was generated
        :return:
        """
        return self._get_where_sql()



"""
The module is there to establish a connection to the database
"""
from typing import Union

import mariadb

from mariadb_sqlbuilder.helpful.validator_dummy import ValidatorDummy
from mariadb_sqlbuilder.helpful.validator import Validator
from mariadb_sqlbuilder.builder import TableBuilder
from mariadb_sqlbuilder.sqlscript.splitter import split_sql_script_in_parameters


class Connector:
    """
    TODO: add a description
    This is a dummy docstring.
    """

    def __init__(self, host: str, user: str, password: str, database: str, *args, port: int = 3306,
                 pool_name: str = "sqlbuilder_pool",
                 pool_size: int = 3, pool_reset_connection: bool = False,
                 use_validator: bool = True, **kwargs):
        self.connections = mariadb.ConnectionPool(
            pool_name=pool_name,
            pool_size=pool_size,
            pool_reset_connection=pool_reset_connection,
            host=host,
            port=port,
            user=user,
            password=password,
            database=database,
            *args,
            **kwargs
        )
        self.__schema = database
        self.connections_list: list = []
        self.in_using_cursors = []
        self.available_cursor: list = []
        self.__is_in_reset = False
        i: int = 1
        while i <= pool_size:
            conn: mariadb.Connection = self.connections.get_connection()
            conn.auto_reconnect = True
            self.connections_list.append(conn)
            self.available_cursor.append(conn.cursor())
            i += 1
        self.__validator = None
        if use_validator:
            self.__validator = Validator(self)
        else:
            self.__validator = ValidatorDummy(self)

    def table(self, name: str) -> TableBuilder:
        """
        Get the functions for building the sql in the selected table.
        :param name: table name
        :return:
        """
        return TableBuilder(self, name)

    def get_available_cursor(self):
        """
        Function for the functionality of execute and fetching.
        The function is used to always give a cursor to an execution
        and let it wait as long as none is available
        :return:
        """
        while not self.available_cursor or self.is_in_reset:
            pass
        cursor = self.available_cursor[0]
        try:
            self.available_cursor.remove(cursor)
        except ValueError:
            cursor = self.get_available_cursor()
        self.in_using_cursors.append(cursor)
        return cursor

    def release_cursor(self, cursor):
        """
        If an execution finish, it releases a cursor and make it
        available for other executions.
        :param cursor:
        :return:
        """
        try:
            self.in_using_cursors.remove(cursor)
        except ValueError:
            cursor.close()
            return
        conn = cursor.connection
        cursor.close()
        cursor = conn.cursor()
        self.available_cursor.append(cursor)

    @property
    def is_in_reset(self) -> bool:
        """
        Get the internal variable if the cursors for
        the execution currently in reset
        :return:
        """
        return self.__is_in_reset

    def reset_available_cursors(self, not_the_save_way: bool = False):
        """
        In case of problems with the cursors, you can reset them.
        :param not_the_save_way: Whether to wait until all actions
        have been performed. Default = False
        :return:
        """
        if not not_the_save_way:
            while len(self.in_using_cursors) > 0:
                pass
        self.__is_in_reset = True
        for cursor in self.available_cursor:
            cursor.close()
        self.available_cursor = []
        self.in_using_cursors = []
        for conn in self.connections_list:
            conn.reset()
        list_of_cursors = []
        for conn in self.connections_list:
            list_of_cursors.append(conn.cursor())
        self.available_cursor = list_of_cursors.copy()
        self.__is_in_reset = False

    def get_active_used_cursors_count(self) -> int:
        """
        Get a count of current used coursers
        :return:
        """
        return len(self.connections_list) - len(self.available_cursor)

    def execute(self, sql: str):
        """
        It will return only if it is successfully,
        only one statement.
        For more statements / a script use 'execute_script'
        :param sql:
        :return:
        """
        cursor = self.get_available_cursor()
        cursor.execute(sql)
        cursor.connection.commit()
        self.release_cursor(cursor)

    def execute_script(self, sql_script: str):
        """
        It will return only if it is complete successfully,
        it will break when a line is not successful.
        :param sql_script:
        :return:
        """
        split_script = split_sql_script_in_parameters(sql_script)
        cursor = self.get_available_cursor()
        for statement in split_script:
            cursor.execute(statement)
        cursor.connection.commit()
        self.release_cursor(cursor)

    def execute_fetchone(self, sql: str):
        """
        Execute a statement with a return of the value
        :param sql:
        :return:
        """
        cursor = self.get_available_cursor()
        cursor.execute(sql)
        result = cursor.fetchone()
        self.release_cursor(cursor)
        return result

    def execute_fetchmany(self, sql: str):
        """
        Execute a statement with a return of the value
        :param sql:
        :return:
        """
        cursor = self.get_available_cursor()
        cursor.execute(sql)
        result = cursor.fetchall()
        self.release_cursor(cursor)
        return result

    @property
    def schema(self) -> str:
        """
        Returns the current schema, where you select your tables
        :return:
        """
        return self.__schema

    def recreate_validator(self):
        """
        Recreate the validator for table changes
        :return:
        """
        self.__validator = Validator(self)

    @property
    def validator(self) -> Union[Validator, ValidatorDummy]:
        """
        Returns the current used validator
        :return:
        """
        return self.__validator



"""
This modul is there to check your database of existing entry's
"""
import mariadb

from .base_builder import ConditionsBuilder


class ExistsBuilder(ConditionsBuilder):
    """
    TODO: add a description
    This is a dummy docstring.
    """

    def __init__(self, tb, **kwargs):
        super().__init__(tb, **kwargs)
        self.column_list = []

    def column(self, column: str):
        """
        Adds a column to the list of columns to select in the EXISTS subquery.
        :param column:
        :return:
        """
        self.column_list.append(column)
        return self

    def columns(self, columns: str):
        """
        Adds multiple columns to the list of columns to select in the EXISTS subquery.
        :param columns:
        :return:
        """
        self.column_list += columns.replace(", ", ",").split(",")
        return self

    def check_exists(self) -> bool:
        """
        Checks if a row exists in the table that matches the specified conditions.
        :return:
        """
        cursor = self.tb.connector.get_available_cursor()
        try:
            cursor.execute(
                self.get_sql()
            )
            result = cursor.fetchone()
        except mariadb.OperationalError as err:
            if "Unknown column" in err.args[0]:
                result = (0,)
            else:
                self.tb.connector.release_cursor(cursor)
                raise mariadb.OperationalError(err)
        self.tb.connector.release_cursor(cursor)
        if result is None:
            return False
        return bool(result[0])

    def get_sql(self) -> str:
        """
        Generates the SQL query string to check for the existence of a row in the table.
        :return:
        """
        if not self.column_list and not self.is_conditions():
            return f"SHOW TABLES LIKE '{self.tb.table}'"
        return f"SELECT EXISTS(SELECT " \
               f"{', '.join(self.column_list) if self.column_list else '*'} " \
               f"FROM {self.tb.table} " \
               f"{self._get_where_sql()};"



"""
This modul is there for build a sql upsert query
"""
from datetime import timedelta, datetime
from json import dumps
from typing import Union, Dict, List

from _decimal import Decimal

from mariadb_sqlbuilder.helpful.arithmetic import Arithmetic
from .base_builder import BaseBuilder, _transform_value_valid


class UpsertBuilder(BaseBuilder):
    """
    TODO: add a description
    This is a dummy docstring.
    """

    def __init__(self, tb, **kwargs):
        super().__init__(tb, **kwargs)
        self.__toSet = {}
        self.__jsonBuildings = []

    def set(self, column: str,  value: Union[str, int, float, bool,
        Arithmetic, timedelta, datetime, Decimal]):
        """
        Set a column value for the current table.
        :param column:
        :param value:
        :return:
        """
        self.tb.validator.check_value_type(self.tb.table, column, value)
        if self.tb.table not in self.__toSet:
            self.__toSet[self.tb.table] = {}
        self.__toSet[self.tb.table][column] = _transform_value_valid(value)
        return self

    def add_join_table(self, table: str):
        """
        Add a join table to the UpsertBuilder object.
        :param table:
        :return:
        """
        if table in self.__toSet:
            return self
        self.__toSet[table] = {}
        return self

    def table_set(self, table: str, column: str, value: Union[str, int, float, bool,
        Arithmetic, timedelta, datetime, Decimal]):
        """
        Set a column value for a specific table.
        :param table:
        :param column:
        :param value:
        :return:
        """
        self.tb.validator.check_value_type(table, column, value)
        if table not in self.__toSet:
            self.__toSet[table] = {}
        self.__toSet[table][column] = _transform_value_valid(value)
        return self

    def execute(self):
        """
        Execute the UpsertBuilder object's SQL query and commit the changes.
        :return:
        """
        cursor = self.tb.connector.get_available_cursor()
        cursor.execute(
            self.get_sql()
        )
        cursor.connection.commit()
        self.tb.connector.release_cursor(cursor)

    def get_sql(self) -> str:
        """
        Get the SQL query string for the UpsertBuilder object.
        :return:
        """
        for element in self.__jsonBuildings:
            self.__set_json(element[0], element[1])
        sql = ""
        _key: str
        _value: Dict[str, dict]
        for _key, _value in self.__toSet.items():
            sql += f"INSERT INTO " \
                   f"{_key} ({', '.join(_value.keys())}) VALUES ({', '.join(_value.values())})" \
                   f"ON DUPLICATE KEY UPDATE " \
                   f"{', '.join([f'{key} = {value}' for (key, value) in _value.items()])};"
        return sql

    def __set_json(self, json: Dict[str, any], pop: List[str] = None):
        """
        Set column values using a JSON object.
        :param json:
        :param pop:
        :return:
        """
        if pop is None:
            pop = []
        key: str
        value: any
        join_keys = list(self.__toSet)
        for key, value in json.items():
            if isinstance(value, dict):
                if key in join_keys and not key in pop:
                    for sub_key, sub_value in value.items():
                        self.table_set(key, sub_key, sub_value)
                else:
                    self.set(key, dumps(value))
            else:
                self.set(key, value)
        return self

    def set_json(self, json: Dict[str, any], pop: List[str] = None):
        """
        Set values with a json.
        :param json: dict with data example from select
        :param pop: pop keys from the json,
        if you have keys inside that are not a table but a dict/list
        :return:
        """
        self.__jsonBuildings.append([json, pop])
        return self



"""
Dummy Class for the Validator,
if you won't use it
"""


class ValidatorDummy:
    """
    This class is for checking the existing of tables, columns and the
    right value types
    """

    def __init__(self, conn):
        self._conn = conn
        self.__structure = {}

    def check_table_exists(self, table: str):
        """
        It's only a dummy function
        :param table:
        :return:
        """

    def check_column_exists(self, table: str, column: str):
        """
        It's only a dummy function
        :param table:
        :param column:
        :return:
        """

    def check_value_type(self, table: str, colum: str, value: any):
        """
        It's only a dummy function
        :param table:
        :param colum:
        :param value:
        :return:
        """

    @property
    def structure(self):
        """
        It's only a dummy function
        :return:
        """
        return self.__structure.copy()



"""
The module is there for a sql script to catch out the statements
"""
import shlex
from typing import List


def split_sql_script_in_parameters(sql_script: str) -> List[str]:
    """
    Split a complete sql statement string in a list of statements
    :param sql_script:
    :return:
    """
    if not sql_script:
        return []

    split_sql = shlex.split(sql_script, posix=False)
    parameters = []
    current_parameters = []
    for x in split_sql:
        if x.startswith(("'", '\'', '"', "\"")):
            current_parameters.append(x)
        elif x.__contains__(";"):
            s_x = x.split(";")
            current_parameters.append(s_x[0] + ";")
            parameters.append(" ".join(current_parameters).strip())
            current_parameters = []
            if s_x[1]:
                current_parameters.append(s_x[1])
        else:
            current_parameters.append(x)
    return parameters



"""
This modul is there for build a sql select query
"""
from typing import Union

from mariadb_sqlbuilder.exepetions import JsonNotSupported
from mariadb_sqlbuilder.helpful.arithmetic import Arithmetic
from .base_builder import ConditionsBuilder, _get_tcn_validator
from .dict_converter import convert_to_dict_single, convert_to_dict_all
from .join_builder import BaseJoinExtension


class SelectBuilder(ConditionsBuilder, BaseJoinExtension):
    """
    TODO: add a description
    This is a dummy docstring.
    """

    def __init__(self, tb, expressions: Union[str, list], *args, **kwargs):
        ConditionsBuilder.__init__(self, tb, **kwargs)
        BaseJoinExtension.__init__(self, tb, **kwargs)
        self._contains_arithmetic = False
        self.expressions = []
        self._loop_tb_expressions_add(self.tb.table, expressions, *args)

    def join_select(self, join_table: str, expressions: Union[str, list], *args):
        """
        Adds a JOIN clause to the SELECT query with the provided table and column.
        :param join_table:
        :param expressions:
        :return:
        """
        self._loop_tb_expressions_add(join_table, expressions, *args)
        return self

    def column_select(self, expressions: Union[str, list], *args):
        """
        Adds additional columns to the SELECT query.
        :param expressions:
        :return:
        """
        self._loop_tb_expressions_add(self.tb.table, expressions, *args)
        return self

    def fetchone(self):
        """
        Executes the SELECT query and returns the first row of the result.
        :return:
        """
        cursor = self.tb.connector.get_available_cursor()
        cursor.execute(
            self.get_sql()
        )
        result = cursor.fetchone()
        cursor.connection.commit()
        self.tb.connector.release_cursor(cursor)
        return result

    def fetchone_json(self):
        """
        Executes the SELECT query and returns the first row of the result in a JSON format.
        :return:
        """
        if self._contains_arithmetic:
            raise JsonNotSupported(
                "This query contains arithmetics, they are not supported in json returns"
            )
        return convert_to_dict_single(self.tb.table, self.expressions, self.fetchone())

    def fetchall(self):
        """
        Executes the SELECT query and returns all the rows of the result.
        :return:
        """
        cursor = self.tb.connector.get_available_cursor()
        cursor.execute(
            self.get_sql()
        )
        result = cursor.fetchall()
        cursor.connection.commit()
        self.tb.connector.release_cursor(cursor)
        return result

    def fetchall_json(self):
        """
        Executes the SELECT query and returns all the rows of the result in a JSON format.
        :return:
        """
        if self._contains_arithmetic:
            raise JsonNotSupported(
                "This query contains arithmetics, they are not supported in json returns"
            )
        return convert_to_dict_all(self.tb.table, self.expressions, self.fetchall())

    def get_sql(self) -> str:
        """
        Builds the SELECT query and returns it as a string.
        :return:
        """
        return f"SELECT {', '.join(self.expressions)} FROM {self.tb.table} " \
               f"{' '.join(self._joins) if self._joins else ''} " \
            f"{self._get_where_sql()};"

    def _loop_tb_expressions_add(self, tb: str, expressions: Union[str, Arithmetic], *args):
        """
        Handle different types of expressions and add it correctly in the expressions list
        :param tb:
        :param expressions:
        :param args:
        :return:
        """
        if isinstance(expressions, str):
            self.expressions += [
                _get_tcn_validator(tb, c, self.tb.validator)
                for c in expressions.replace(", ", ",").split(",")
            ]
        if isinstance(expressions, list) or args:
            loop_expressions = []
            if isinstance(expressions, list):
                loop_expressions += expressions
            if args:
                loop_expressions += args
            for expression in loop_expressions:
                if isinstance(expression, Arithmetic):
                    self._contains_arithmetic = True
                    self.expressions.append(str(expression))
                else:
                    self.expressions.append(_get_tcn_validator(tb, expression, self.tb.validator))



"""
This module is there for parse sql fetches to a dict
"""
from typing import List, Dict


def __loop_convert_to_dict(mtb: str, columns: List[List[str]], values: List[any]) -> Dict[str, any]:
    result: Dict[str, any] = {}
    value: any
    for column, value in zip(columns, values):
        if column[1] == "*":
            raise TypeError('Column * is not supported in this dict/json')
        if column[0] != mtb:
            if column[0] not in result:
                result[column[0]] = {}

            result[column[0]][column[1]] = value
        else:
            result[column[1]] = value
    return result


def convert_to_dict_single(mtb: str, columns: List[str], values: List[any]) -> Dict[str, any]:
    """
    Convert a single result from fetch to a dict/json
    :param mtb: main table
    :param columns: list of all columns
    :param values:  list of the result
    :return: result as json
    """
    columns: List[List[str]] = [column.split(".") for column in columns]
    result: Dict[str, any] = __loop_convert_to_dict(mtb, columns, values)

    return result


def convert_to_dict_all(mtb: str, columns: List[str],
                        values: List[List[any]]) -> List[Dict[str, any]]:
    """
    Convert a multiple result to a list with the results as dict/json
    :param mtb: main table
    :param columns: list of all columns
    :param values: list with the list of results
    :return:
    """
    result: List[Dict[str, any]] = []
    # split columns, better performance
    column: str
    columns: List[List[str]] = [column.split(".") for column in columns]
    # define types
    v: List[any]
    column: List[str]
    for v in values:
        result.append(__loop_convert_to_dict(mtb, columns, v))
    return result



"""
The module is there to give the possibility to access
filters without having to enter them separately
"""
EQUALS = "="
NOT_EQUALS = "!="
GREATER = ">"
GREATER_EQUALS = ">="
LESS = "<"
LESS_EQUALS = "<="



"""
This modul is there for build a sql delete query for
delete stuff in your database
"""
from .base_builder import ConditionsBuilder


class DeleteBuilder(ConditionsBuilder):
    """
    TODO: add a description
    This is a dummy docstring.
    """

    def __init__(self, tb, **kwargs):
        super().__init__(tb, **kwargs)
        self.sure_not_use_conditions = False

    def im_sure_im_not_use_conditions(self, im_sure: bool = True):
        """
        Sets the sure_not_use_conditions attribute to the provided boolean value.
        :param im_sure:
        :return:
        """
        self.sure_not_use_conditions = im_sure
        return self

    def execute(self):
        """
        Executes the DELETE SQL statement built by the builder.
        :return:
        """
        if not self.is_conditions() and not self.sure_not_use_conditions:
            raise PermissionError('Delete Builder: You are not sure enough not to use where')
        cursor = self.tb.connector.get_available_cursor()
        cursor.execute(
            self.get_sql()
        )
        cursor.connection.commit()
        self.tb.connector.release_cursor(cursor)

    def get_sql(self) -> str:
        """
        Returns the DELETE SQL statement built by the builder as a string.
        :return:
        """
        return f"DELETE FROM {self.tb.table} " \
            f"{self._get_where_sql()};"



"""
All custom exceptions for the sql builder
"""


class BetweenValueIsBigger(Exception):
    """
    Raises if value 1 is bigger than value 2,
    usually value 1 need to be smaller or the same then value 2
    """


class JsonNotSupported(Exception):
    """
    If Json not supported, in example if you use Arithmetic, there's
    no name for that key
    """


class InvalidColumnType(Exception):
    """
    If a type of Column not supported in the validator system
    """


class ValidatorType(Exception):
    """
    The python type of the given value are not a supported type
    of the column
    """


class ValidatorLength(Exception):
    """
    The length of the value is not in the allowed length of the column
    """


class ValidatorRange(Exception):
    """
    If a Number not in the allowed range
    """


class ValidatorUnknown(Exception):
    """
    If is an error unexpectedly, because it shouldn't be possible
    """


class ValidatorTableNotFound(Exception):
    """
    If a table not registered in the validator system
    """


class ValidatorColumnNotFound(Exception):
    """
    If a Column of a table not registered in the validator system
    """



import pathlib
from setuptools import setup

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

VERSION = '1.0.0a6'

setup(
    name='MariaDB-SQLBuilder',
    long_description=README,
    long_description_content_type="text/markdown",
    version=VERSION,
    packages=[
        "mariadb_sqlbuilder",
        "mariadb_sqlbuilder.builder",
        "mariadb_sqlbuilder.sqlscript",
        "mariadb_sqlbuilder.helpful",
    ],
    url='https://github.com/princessmiku/MariaDB-SQLBuilder',
    license='LGPL 2.1',
    author='Miku',
    author_email='',
    description='MariaDB SQL Builder is a simple way to use Maria SQL. '
                'Use your own SQL or use the integrated Maria SQL Builder tool.',
    keywords=['database', 'mariadb', 'sql', 'builder', 'script builder', 'mariadb sql', 'orm'],
    python_requires='>=3.7.0',
    install_requires=[
        "mariadb>=1.1.4",
    ],
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
)



"""
This modul is there for handle table functions
"""
from typing import Union

from mariadb_sqlbuilder.helpful.validator import Validator
from .select_builder import SelectBuilder
from .update_builder import UpdateBuilder
from .insert_builder import InsertBuilder
from .upsert_builder import UpsertBuilder
from .delete_builder import DeleteBuilder
from .exists_builder import ExistsBuilder


class TableBuilder:
    """
    TODO: add a description
    This is a dummy docstring.
    """

    def __init__(self, connector, table: str):
        connector.validator.check_table_exists(table)
        self.__connector = connector
        self.__table = table

    def select(self, expressions: Union[str, list], *args, **kwargs) -> SelectBuilder:
        """
        Returns a SelectBuilder instance to build a SELECT query.
        :param expressions:
        :param kwargs:
        :return:
        """
        return SelectBuilder(self, expressions, *args, **kwargs)

    def update(self, **kwargs) -> UpdateBuilder:
        """
        Returns an UpdateBuilder instance to build an UPDATE query.
        :param kwargs:
        :return:
        """
        return UpdateBuilder(self, **kwargs)

    def insert(self, **kwargs) -> InsertBuilder:
        """
        Returns an InsertBuilder instance to build an INSERT query.
        :param kwargs:
        :return:
        """
        return InsertBuilder(self, **kwargs)

    def upsert(self, **kwargs) -> UpsertBuilder:
        """
        Returns an UpsertBuilder instance to build an UPSERT query.
        :param kwargs:
        :return:
        """
        return UpsertBuilder(self, **kwargs)

    def delete(self, **kwargs) -> DeleteBuilder:
        """
        Returns a DeleteBuilder instance to build a DELETE query.
        :param kwargs:
        :return:
        """
        return DeleteBuilder(self, **kwargs)

    def exists(self, **kwargs) -> ExistsBuilder:
        """
        Returns an ExistsBuilder instance to check if a record exists in the table.
        :param kwargs:
        :return:
        """
        return ExistsBuilder(self, **kwargs)

    @property
    def table(self):
        """
        Protect the table for changes
        :return:
        """
        return self.__table

    @property
    def validator(self) -> Validator:
        """
        Returns the current validator
        :return:
        """
        return self.__connector.validator

    @property
    def connector(self):
        """
        Protect the connector for changes
        :return:
        """
        return self.__connector
